# Stores and retrieves cookies using a hash value.
# Designed to handle persistence even in a multiple server environment.
# Copyright (C) 2000 Shane Hathaway
# E-mail: shathaway@earthling.net
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

__doc__ = '''BasicCookieJar class.'''
__version__ = '0.1'

import sha, thread, time, whrandom
from urllib import unquote
from Globals import Persistent

_generator = whrandom.whrandom()

class Cookie:
    def __init__(self, cookieName, hash, internalValue,
                 expiration, timeoutSeconds):
        self.cookieName = cookieName
        self.hash = hash
        self.internalValue = internalValue
        self.expiration = expiration
        self.timeoutSeconds = timeoutSeconds


class BasicCookieJar (Persistent):
    '''Stores and retrieves cookies having an arbitrary Python
    object value.  The hash value sent to the browser
    is completely unpredictable, making it secure when used
    the right way.  The persistent storage could be shared among
    different Zope servers, making it possible to recognize
    cookies on a super-site as well as retain values when Zope
    restarts.'''

    def __init__(self, cleanupInterval):
        self.cleanupInterval = cleanupInterval
        self.__prepareVars()

    def __prepareVars(self):
        '''Prepares the non-persistent variables.'''
        if not hasattr(self, "_v_hashtable"):
            self._v_hashtable = {} # A dictionary of dictionaries.
            self._v_nextCleanup = 0
            self._v_cleanupThreadLock = thread.allocate_lock()

    def setCookie(self, REQUEST, cookieName,
                  internalValue, timeoutSeconds):
        '''Sets a cookie as a hash value which can be looked up
        in future requests by getCookie().'''
        self.__prepareVars()
        now = time.time()
        self.__mayCleanup(now)
        if self.getCookie(REQUEST, cookieName) != internalValue:
            # Produce a random value then scramble it, making it
            # virtually impossible to predict hash values.
            # Also the hash value must be large so as to make hashes
            # virtually impossible to duplicate or predict.
            shaVal = sha.new("%f%f" % (now, _generator.random()))
            hash = shaVal.hexdigest()[4:28]  # Not too large a cookie name.
            # If the name is too large, the persistent storage might
            # reject it.
            expiration = now + timeoutSeconds
            ck = Cookie(cookieName, hash, internalValue,
                        expiration, timeoutSeconds)
            self._storeCookieInHashtable(ck)
            self._setPersistentCookie(ck)
            REQUEST.RESPONSE.setCookie(cookieName, hash, path='/')

    def getCookie(self, REQUEST, cookieName):
        '''Retrieves the internal value of a cookie.'''
        self.__prepareVars()
        localExpiration = 0
        if REQUEST.has_key(cookieName):
            hash = unquote(REQUEST[cookieName])
            group = self.__accessCookieGroup(cookieName, create=0)
            if group is not None and group.has_key(hash):
                try:
                    ck = group[hash]
                    now = time.time()
                    # print ck.expiration, now, ck.timeoutSeconds
                    if ck.expiration > now:
                        # Update the expiration time.
                        ck.expiration = now + ck.timeoutSeconds
                        return ck.internalValue
                    else:
                        # Expired here.
                        localExpiration = 1
                        del group[hash]
                except KeyError:
                    # The cookie was just cleaned up by another thread.
                    pass
            # See if the cookie is in persistent storage.
            ck = self._getPersistentCookie(cookieName, hash)
            if ck is not None:
                if localExpiration and ck.expiration <= now:
                    # This cookie just expired here as well as
                    # in persistent storage.  Kill it for good.
                    self._deletePersistentCookie(cookieName, hash)
                    return None
                now = time.time()
                ck.expiration = now + ck.timeoutSeconds
                self._storeCookieInHashtable(ck)
                return ck.internalValue
        return None

    def _obscureCookie(self, REQUEST, cookieName):
        '''Makes it impossible to retrieve the cookie in DTML.'''
        blankStr = 'This space intentionally left blank.'
        REQUEST.form[cookieName] = blankStr
        REQUEST[cookieName] = blankStr

    def listCookies(self, REQUEST):
        '''Returns a dictionary of values of cookies in this request.'''
        self.__prepareVars()
        now = time.time()
        self.__mayCleanup(now)

        rval = []
        
        for cookieName, group in self._v_hashtable.items():
            if REQUEST.has_key(cookieName):
                hash = unquote(REQUEST[cookieName])
                if group.has_key(hash):
                    try:
                        ck = group[hash]
                        if ck.expiration > now:
                            rval.append(ck)
                        else:
                            del group[hash]
                    except KeyError:
                        # Cookie was removed just now.
                        pass
        return rval

    def deleteCookie(self, REQUEST, cookieName):
        '''Deletes a cookie value from the table for the purpose of logging
        out or debugging.
        '''
        if REQUEST.has_key(cookieName):
            hash = unquote(REQUEST[cookieName])
            group = self.__accessCookieGroup(cookieName, create=0)
            try:
                del group[hash]
            except:
                # ignore.
                pass
            self._deletePersistentCookie(cookieName, hash)

    def changeProperties(self, cleanupInterval):
        '''Changes the properties of this CookieJar.'''
        self.cleanupInterval = cleanupInterval

    def _storeCookieInHashtable(self, ck):
        self._v_cleanupThreadLock.acquire() # Disallow cleanup for a moment.
        # Otherwise the cookie might be stored in a group that's about
        # to disappear.
        try:
            group = self.__accessCookieGroup(ck.cookieName, create=1)
            group[ck.hash] = ck
        finally:
            self._v_cleanupThreadLock.release()

    def __accessCookieGroup(self, cookieName, create):
        '''Accesses the section of _v_hashtable that stores info for
        cookies by this name. May create if necessary.'''
        if self._v_hashtable.has_key(cookieName):
            try:
                return self._v_hashtable[cookieName]
            except KeyError:
                # The group was just removed by another thread.
                pass
        if create:
            # Create a new entry in the top-level hashtable.
            dict = {}
            self._v_hashtable[cookieName] = dict
            return dict
        else:
            return None

    def __mayCleanup(self, now):
        '''Clean up cookies if necessary.'''
        if self.cleanupInterval > 0 and self._v_nextCleanup <= now:
            self._v_nextCleanup = now + self.cleanupInterval
            # Only let one thread perform the cleanup.
            if self._v_cleanupThreadLock.acquire(0): # Don't wait.
                try:
                    now = time.time()
                    for cookieName, group in self._v_hashtable.items():
                        for hash, ck in group.items():
                            if ck.expiration <= now:
                                del group[hash]
                        if len(group) < 1:
                            # No more cookies in this category remain.
                            del self._v_hashtable[cookieName]
                finally:
                    self._v_cleanupThreadLock.release()
                # Now have the persistent storage clean up too.
                self._cleanupPersistentStorage()

    def _getInternalExpiration(self, cookieName, hash):
        '''This is provided to allow the persistent storage mechanism
        to discover what to do with a given persistent cookie. Returns
        0 if nonexistent.'''
        group = self.__accessCookieGroup(cookieName, create=0)
        if group is not None and group.has_key(hash):
            try:
                ck = group[hash]
                return ck.expiration
            except KeyError:
                # Cookie was just cleaned up by another thread.
                return 0
        return 0
                    
    def _cleanupPersistentStorage(self):
        '''Abstract method.  Removes expired cookies from the persistent
        storage area.'''
        pass

    def _getPersistentCookie(self, cookieName, hash):
        '''Abstract method.  Retrieves a Cookie object.'''
        pass

    def _setPersistentCookie(self, ck):
        '''Abstract method.  Stores a cookie value.'''
        pass
        
    def _deletePersistentCookie(self, cookieName, hash):
        '''Abstract method.'''
        pass
